/***************************************************
TASS Control Software by John Ackermann N8UR.  Based on Adafruit examples below
****************************************************/

char version[] = "v0.45 -- 16 Dec 2015";

// BOF preprocessor bug prevent - insert me on top of your arduino-code
// From: http://www.a-control.de/arduino-fehler/?lang=en
#if 1
__asm volatile ("nop");
#endif

// uncomment to enable Adafruit 2.8 inch TFT and capacitive touchscreen
// #define USE_TS

// uncomment to endable Adafruit 1.2 inch 8x8 LED matrix
#define USE_LED8X8

// uncomment to enable Wiznet W5100 ethernet shield
// #define USE_W5100

// comment this to set board up as single 8 pole using K9
// #define SPLIT

// uncomment this to display bitmasks on the serial console
#define SHOW_MASKS

#include "TASS_defines.h"

#ifdef USE_TS
  #include <Wire.h>
  #include <SPI.h>
  #include <Adafruit_GFX.h>
  #include <Adafruit_ILI9341.h>
  #include <Adafruit_FT6206.h>
  Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC);
  Adafruit_FT6206 ctp = Adafruit_FT6206();
#endif // USE_TS

#ifdef USE_LED8X8
  #include <Wire.h>
  #include <SPI.h>
  #include <Adafruit_GFX.h>
  #include <Adafruit_LEDBackpack.h>
  Adafruit_8x8matrix matrix = Adafruit_8x8matrix();
  #define HB_BLINK_MS 1000;
 #endif // USE_LED8X8 

#ifdef USE_W5100
  #include <SPI.h>
  #include <EthernetUdp.h>
  #include <Dhcp.h>
  #include <Dns.h>
  #include <util.h>
  #include <Ethernet.h>
  #include <EthernetServer.h>
  #include <EthernetClient.h>

// ******************************************************
// If using ethernet, set network stuff here:
  byte mac[] = { 0xDE, 0xAD, 0xBE, 0xEF, 0xFE, 0xED };
  IPAddress ip(10,73,3,177);
  IPAddress gateway(10,73,1,1);
  IPAddress subnet(255,255,0,0);
  
  // what to echo to the telnet port, if anything
  // 0 = no echo, 1 = echo raw command, 2 = interpreted output
  int telnet_verbose = 2; 
//*******************************************************
  
  // variable
 EthernetServer server = EthernetServer(23);
 EthernetClient client = 0;
 
 #endif  // USE_W5100

char board = ' ';
int relay = 0;
char cmd = ' ';
byte relay_num = 0;
char relay_char = ' '; 
boolean newclient = 1;
boolean execute = 0;
boolean BadCmd = 0;

// these keep the relay settings for each board across commands
byte S_mask_A = 0;
byte S_mask_B = 0;
byte S_mask_C = 0;
byte S_mask_D = 0;

// these keep the hold settings for each board across commands
byte H_mask_A = 0;
byte H_mask_B = 0;
byte H_mask_C = 0;
byte H_mask_D = 0;

// these hold the current status (set + hold) for each board
byte cmd_mask_A = 0;
byte cmd_mask_B = 0;
byte cmd_mask_C = 0;
byte cmd_mask_D = 0;

// setup for serial input
const char startChar = '<';
const char endChar = '>';
char dataBuffer[DATABUFFERSIZE+1]; //Add 1 for NULL terminator

void setup(void) {
// set configuration pins to input pullup
  pinMode(14,INPUT_PULLUP);
  pinMode(15,INPUT_PULLUP);
  pinMode(16,INPUT_PULLUP);
  pinMode(17,INPUT_PULLUP);
  pinMode(18,INPUT_PULLUP);
  pinMode(19,INPUT_PULLUP);
  pinMode(22,INPUT_PULLUP);         
  
// initialize serial:
  Serial.begin(115200);
// splash
  Serial.println("TAPR TASS Control Software");
  Serial.println(version);
  Serial.println("Copyright 2015 John Ackermann   N8UR");
  Serial.println("Licensed under MIT license");
  Serial.println("");

// set up digital pins for each board
  for (int relay_num=A_1;relay_num<=A_1+8;relay_num++){
    pinMode(relay_num, OUTPUT);}
  for (int relay_num=B_1;relay_num<=B_1+8;relay_num++){
    pinMode(relay_num, OUTPUT);}
  for (int relay_num=C_1;relay_num<=C_1+8;relay_num++){
    pinMode(relay_num, OUTPUT);}
  for (int relay_num=D_1;relay_num<=D_1+8;relay_num++){
    pinMode(relay_num, OUTPUT);}
  
#ifdef USE_TS
  tft_setup();
#endif

#ifdef USE_LED8X8
  matrix.begin(0x70);
  matrix.clear();
  matrix.setBrightness(0);
  matrix.drawPixel(0,0,LED_ON);
  matrix.writeDisplay();
  delay(1000);
  int col;
  int row;
  for( col = 0; col <= 7; col++ ){
    for( row = 0; row <= 7; row++ ){
      matrix.drawPixel(row,col,LED_ON);
      matrix.writeDisplay();
      delay(10);
    }
  }
  delay(250);
  matrix.clear();
  matrix.writeDisplay();
  
#endif

#ifdef USE_W5100
  //disable SD
  pinMode(4,OUTPUT);
  digitalWrite(4,HIGH);

  Ethernet.begin(mac, ip, gateway, subnet);
  server.begin();    
#endif

} // setup

void loop() {
  board = ' ';
  relay = 0;
  cmd = ' ';
  BadCmd = 0;
  
  heartbeat();

  if (!execute) {
    #ifdef USE_TS 
      get_ts_data();
    #endif
    #ifdef USE_W5100
      get_w5100_data();
    #endif
    get_usb_data();
 }
 if (execute) {
    board = toupper(dataBuffer[1]);
    relay = atoi(&dataBuffer[2]);
    cmd = toupper(dataBuffer[3]); 
    
    byte cmd_mask = 0;
      
    // if S set the relay mask
    if (cmd == 'S') {
      
    // first, clear mask
     switch (board) {
      case 'A': S_mask_A = clear_S_mask(relay,S_mask_A);break;
      case 'B': S_mask_B = clear_S_mask(relay,S_mask_B);break;
      case 'C': S_mask_C = clear_S_mask(relay,S_mask_C);break;
      case 'D': S_mask_D = clear_S_mask(relay,S_mask_D);break;
        } // switch
    
   // then set the relay mask
   switch (board) {
      case 'A': S_mask_A = set_mask(relay,S_mask_A); break;
      case 'B': S_mask_B = set_mask(relay,S_mask_B); break;
      case 'C': S_mask_C = set_mask(relay,S_mask_C); break;
      case 'D': S_mask_D = set_mask(relay,S_mask_D); break;
        } // switch
    }
    
    // if U clear appropriate hold mask
    if (cmd == 'U') {
      
      // special case - relay 0 means clear everything
      // don't use "clear_set_mask" because we want to wipe both halves
      if (relay == 0) { 
        switch(board) {
          case 'A': S_mask_A = clear_mask(S_mask_A);break;
          case 'B': S_mask_B = clear_mask(S_mask_B);break;
          case 'C': S_mask_C = clear_mask(S_mask_C);break;
          case 'D': S_mask_D = clear_mask(S_mask_D);break;
        } // switch
      } // if
    
      switch (board) {
        case 'A': H_mask_A = clear_mask(H_mask_A); break;
        case 'B': H_mask_B = clear_mask(H_mask_B); break;
        case 'C': H_mask_C = clear_mask(H_mask_C); break;
        case 'D': H_mask_D = clear_mask(H_mask_D); break;
      } // switch
   } // if

  // if H set appropriate hold mask
    if (cmd == 'H') {
      switch (board) {
        case 'A': H_mask_A = set_mask(relay,H_mask_A); break;
        case 'B': H_mask_B = set_mask(relay,H_mask_B); break;
        case 'C': H_mask_C = set_mask(relay,H_mask_C); break;
        case 'D': H_mask_D = set_mask(relay,H_mask_D);break;
        } // switch
      } // if
      
    cmd_mask_A = H_mask_A | S_mask_A;
    cmd_mask_B = H_mask_B | S_mask_B;
    cmd_mask_C = H_mask_C | S_mask_C;
    cmd_mask_D = H_mask_D | S_mask_D;

  // write to appropriate board
    switch (board) {
      case 'A': write_board(A_1,cmd_mask_A);break;
      case 'B': write_board(B_1,cmd_mask_B);break;
      case 'C': write_board(C_1,cmd_mask_C);break;
      case 'D': write_board(D_1,cmd_mask_D);break;
  } // switch

  print_status();
  
  #ifdef USE_TS
    draw_status();
  #endif
  
  #ifdef USE_LED8X8
    draw_LED_status();
  #endif
  
  execute = 0;
  } // execute 

} // of loop
  
// #################################

byte set_mask(int x, byte mask) {
  switch (x) {
    case 1: mask |= 0x01; break;
    case 2: mask |= 0x02; break;
    case 3: mask |= 0x04; break;
    case 4: mask |= 0x08; break;
    case 5: mask |= 0x10; break;
    case 6: mask |= 0x20; break;
    case 7: mask |= 0x40; break;
    case 8: mask |= 0x80; break;
  }
  return mask;
} // of set_mask

byte clear_mask_top(byte mask) {
  mask = mask & 0xF0;
     return mask;   
}

byte clear_mask_bottom(byte mask) {
  mask = mask & 0x0F;
  return mask;  
}

byte clear_mask(byte mask) { 
  mask = mask & 0x00;
  return mask;  
}

byte clear_S_mask(int relay,byte mask) {
 #ifdef SPLIT
  if (relay >=1 && relay <=4) { 
    mask = clear_mask_top(mask);
    } else {
    mask = clear_mask_bottom(mask);
  } // if   
  #else
  mask = clear_mask(mask);
  #endif
  return mask;  
}

void write_board(int first_io, byte cmd_mask) {
    int counter = first_io; // first relay
    for (byte mask = 00000001; mask>0; mask <<= 1) { //iterate through bit mask
      if (cmd_mask & mask) { // if bitwise AND resolves to true
        digitalWrite(counter,HIGH);
        } // of if
      counter++;
    } // of for

  #ifndef SPLIT
  // set K9
  if (relay >4) {
    digitalWrite(first_io+8,HIGH);
  } else {
    digitalWrite(first_io+8,LOW);
  }  // if
  #endif
}  // of write_board

void print_status(){
  Serial.println("");
  Serial.print ("Board: ");
  Serial.print (board);
  Serial.print (" relay: ");
  Serial.print (relay);
  Serial.print (" command: ");
  Serial.print (cmd);

  if ((cmd == 'U') && (relay == 0)) { Serial.print(" -- Clear All"); }
  Serial.println();
  
  #ifdef SHOW_MASKS
    display_masks();
    Serial.println();
  #endif
} // print_status()

void display_masks() {
  
   Serial.print("\n");
   Serial.print(" S Masks:     ");
   printBits(S_mask_A); 
   Serial.print(" ");
   printBits(S_mask_B); 
   Serial.print(" ");
   printBits(S_mask_C); 
   Serial.print(" ");
   printBits(S_mask_D); 
   Serial.print("\n");
   Serial.print(" H Masks:     ");
   printBits(H_mask_A); 
   Serial.print(" ");
   printBits(H_mask_B); 
   Serial.print(" ");
   printBits(H_mask_C); 
   Serial.print(" ");
   printBits(H_mask_D); 
   Serial.print("\n");
   Serial.print(" Write Masks: ");
   printBits(cmd_mask_A); 
   Serial.print(" ");
   printBits(cmd_mask_B); 
   Serial.print(" ");
   printBits(cmd_mask_C); 
   Serial.print(" ");
   printBits(cmd_mask_D); 
   Serial.print("\n");
} // display_masks()
   
void printBits(byte myByte){
 for (byte mask = 0x80; mask; mask >>= 1){
   if (mask  & myByte)
       Serial.print('1');
   else
       Serial.print('0');
   }
} // printBits()

void draw_LED_status(){
  set_LED_row(2,cmd_mask_A);
  set_LED_row(3,cmd_mask_B);
  set_LED_row(4,cmd_mask_C);
  set_LED_row(5,cmd_mask_D);
} // draw_LED_status()

void set_LED_row(int row, byte myByte){
  int col = 7;  // count backwards since bit mask moves LSB -> MSB
  for (byte mask = 0x80; mask; mask >>= 1){
    matrix.drawPixel(col,row,LED_OFF); // clear previous
    if (mask  & myByte) {
      matrix.drawPixel(col,row,LED_ON);
    }
    col--; 
  }
  matrix.writeDisplay();
} // set_LED_row

void heartbeat() {
  // based on LED Heartbeat routine by Allen C. Huffman (www.appleause.com)
  static unsigned int  hbStatus = LOW;  // Last set mode.
  static unsigned long hbBlinkTime = 0; // blink time.

  // LED blinking heartbeat. Yes, we are alive.
  // For explanation, see:
  // http://playground.arduino.cc/Code/TimingRollover
  if ( (long)(millis()-hbBlinkTime) >= 0 )
  {
    // Toggle LED.
    hbStatus = (hbStatus==HIGH ? LOW : HIGH);
    matrix.drawPixel(7,7,hbStatus);
    matrix.writeDisplay();

    // Reset "next time to toggle" time.
    hbBlinkTime = millis()+HB_BLINK_MS;
  }
} // End of ledBlink()



